@@ -19,9 +19,7 @@ class BatchingConfiguration(ABC):
     @property
     @abstractmethod
     def batch_size(self) -> int:
-        """
-        Number of messages to process at once if batching_enabled
-        """
+        """Number of messages to process at once if batching_enabled"""
         ...
 
 
@@ -40,7 +38,7 @@ class NoBatching(BatchingConfiguration):
 class BatchMessages(BatchingConfiguration):
     """Configures the processor to send a list of messages to the call handler"""
 
-    def __init__(self, batch_size):
+    def __init__(self, batch_size: int):
         self.number_of_messages = batch_size
 
     @property
