@@ -1,11 +1,25 @@
+import json
 import logging
 import uuid
 from collections import defaultdict
 from contextlib import contextmanager
 from functools import partial
-from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    Generator,
+    Iterable,
+    List,
+    Literal,
+    Optional,
+    Tuple,
+    TypeVar,
+)
 
 import attr
+from typing_extensions import ParamSpec
 
 from sqs_workers import DEFAULT_BACKOFF, codecs
 from sqs_workers.async_task import AsyncTask
@@ -15,9 +29,11 @@ from sqs_workers.core import BatchProcessingResult, get_job_name
 from sqs_workers.exceptions import SQSError
 from sqs_workers.processors import DEFAULT_CONTEXT_VAR, Processor
 from sqs_workers.shutdown_policies import NEVER_SHUTDOWN
+from sqs_workers.utils import batcher
 
 DEFAULT_MESSAGE_GROUP_ID = "default"
 SEND_BATCH_SIZE = 10
+MAX_MESSAGE_LENGTH = 262144  # 256 KiB
 
 if TYPE_CHECKING:
     from sqs_workers import SQSEnv
@@ -25,13 +41,17 @@ if TYPE_CHECKING:
 logger = logging.getLogger(__name__)
 
 
+P = ParamSpec("P")
+R = TypeVar("R")
+
+
 class SQSBatchError(SQSError):
     def __init__(self, errors):
         self.errors = errors
 
 
 @attr.s
-class GenericQueue(object):
+class GenericQueue:
     env: "SQSEnv" = attr.ib(repr=False)
     name: str = attr.ib()
     backoff_policy: BackoffPolicy = attr.ib(default=DEFAULT_BACKOFF)
@@ -43,9 +63,7 @@ class GenericQueue(object):
         return partial(cls, **kwargs)
 
     def process_queue(self, shutdown_policy=NEVER_SHUTDOWN, wait_second=10):
-        """
-        Run worker to process one queue in the infinite loop
-        """
+        """Run worker to process one queue in the infinite loop"""
         logger.debug(
             "Start processing queue %s",
             self.name,
@@ -70,53 +88,84 @@ class GenericQueue(object):
                 )
                 break
 
-    def process_batch(self, wait_seconds=0) -> BatchProcessingResult:
+    def process_batch(self, wait_seconds: int = 0) -> BatchProcessingResult:
         """
         Process a batch of messages from the queue (10 messages at most), return
         the number of successfully processed messages, and exit
         """
+        if self.batching_policy.batching_enabled:
+            messages = self.get_raw_messages(
+                wait_seconds, self.batching_policy.batch_size
+            )
+            success = self.process_messages(messages)
+            messages_with_success = ((m, success) for m in messages)
+        else:
+            messages = self.get_raw_messages(wait_seconds)
+            success = [self.process_message(message) for message in messages]
+            messages_with_success = zip(messages, success)
+
+        return self._handle_processed(messages_with_success)
+
+    def _handle_processed(self, messages_with_success: Iterable[Tuple[Any, bool]]):
+        """
+        Handles the results of processing messages.
+
+        For successful messages, we delete the message ID from the queue, which is
+        equivalent to acknowledging it.
+
+        For failed messages, we change the visibility of the message, in order to
+        keep it un-consumeable for a little while (a form of backoff).
+
+        In each case (delete or change-viz), we batch the API calls to AWS in order
+        to try to avoid getting throttled, with batches of size 10 (the limit). The
+        config (see sqs_env.py) should also retry in the event of exceptions.
+        """
         queue = self.get_queue()
 
-        if self.batching_policy.batching_enabled:
-            return self._process_messages_in_batch(queue, wait_seconds)
-
-        return self._process_messages_individually(queue, wait_seconds)
-
-    def _process_messages_in_batch(self, queue, wait_seconds):
-        messages = self.get_raw_messages(wait_seconds, self.batching_policy.batch_size)
         result = BatchProcessingResult(self.name)
 
-        success = self.process_messages(messages)
+        for subgroup in batcher(messages_with_success, batch_size=10):
+            entries_to_ack = []
+            entries_to_change_viz = []
 
-        for message in messages:
-            result.update_with_message(message, success)
-            if success:
-                entry = {
-                    "Id": message.message_id,
-                    "ReceiptHandle": message.receipt_handle,
-                }
-                queue.delete_messages(Entries=[entry])
-            else:
-                timeout = self.backoff_policy.get_visibility_timeout(message)
-                message.change_visibility(VisibilityTimeout=timeout)
-        return result
+            for m, success in subgroup:
+                result.update_with_message(m, success)
+                if success:
+                    entries_to_ack.append(
+                        {
+                            "Id": m.message_id,
+                            "ReceiptHandle": m.receipt_handle,
+                        }
+                    )
+                else:
+                    entries_to_change_viz.append(
+                        {
+                            "Id": m.message_id,
+                            "ReceiptHandle": m.receipt_handle,
+                            "VisibilityTimeout": self.backoff_policy.get_visibility_timeout(
+                                m
+                            ),
+                        }
+                    )
 
-    def _process_messages_individually(self, queue, wait_seconds):
-        messages = self.get_raw_messages(wait_seconds)
-        result = BatchProcessingResult(self.name)
+            ack_response = queue.delete_messages(Entries=entries_to_ack)
+
+            if ack_response.get("Failed"):
+                logger.warning(
+                    "Failed to delete processed messages from queue",
+                    extra={"queue": self.name, "failures": ack_response["Failed"]},
+                )
+
+            viz_response = queue.change_message_visibility_batch(
+                Entries=entries_to_change_viz,
+            )
+
+            if viz_response.get("Failed"):
+                logger.warning(
+                    "Failed to change visibility of messages which failed to process",
+                    extra={"queue": self.name, "failures": viz_response["Failed"]},
+                )
 
-        for message in messages:
-            success = self.process_message(message)
-            result.update_with_message(message, success)
-            if success:
-                entry = {
-                    "Id": message.message_id,
-                    "ReceiptHandle": message.receipt_handle,
-                }
-                queue.delete_messages(Entries=[entry])
-            else:
-                timeout = self.backoff_policy.get_visibility_timeout(message)
-                message.change_visibility(VisibilityTimeout=timeout)
         return result
 
     def process_message(self, message: Any) -> bool:
@@ -136,17 +185,16 @@ class GenericQueue(object):
         """
         raise NotImplementedError()
 
-    def get_raw_messages(self, wait_seconds, max_messages=10):
-        """
-        Return raw messages from the queue, addressed by its name
-        """
+    def get_raw_messages(self, wait_seconds: int, max_messages: int = 10) -> List[Any]:
+        """Return raw messages from the queue, addressed by its name"""
+        queue = self.get_queue()
+
         kwargs = {
             "WaitTimeSeconds": wait_seconds,
             "MaxNumberOfMessages": max_messages if max_messages <= 10 else 10,
             "MessageAttributeNames": ["All"],
             "AttributeNames": ["All"],
         }
-        queue = self.get_queue()
 
         if max_messages <= 10:
             return queue.receive_messages(**kwargs)
@@ -165,24 +213,35 @@ class GenericQueue(object):
         return received_messages
 
     def drain_queue(self, wait_seconds=0):
-        """
-        Delete all messages from the queue without calling purge().
-        """
+        """Delete all messages from the queue without calling purge()."""
         queue = self.get_queue()
+
         deleted_count = 0
         while True:
             messages = self.get_raw_messages(wait_seconds)
             if not messages:
                 break
+
             entries = [
                 {"Id": msg.message_id, "ReceiptHandle": msg.receipt_handle}
                 for msg in messages
             ]
-            queue.delete_messages(Entries=entries)
+
+            ack_response = queue.delete_messages(Entries=entries)
+
+            if ack_response.get("Failed"):
+                logger.warning(
+                    "Failed to delete processed messages from queue",
+                    extra={
+                        "queue": self.name,
+                        "failures": ack_response["Failed"],
+                    },
+                )
+
             deleted_count += len(messages)
         return deleted_count
 
-    def get_sqs_queue_name(self):
+    def get_sqs_queue_name(self) -> str:
         """
         Take "high-level" (user-visible) queue name and return SQS
         ("low level") name by simply prefixing it. Used to create namespaces
@@ -191,9 +250,7 @@ class GenericQueue(object):
         return self.env.get_sqs_queue_name(self.name)
 
     def get_queue(self):
-        """
-        Helper function to return queue object.
-        """
+        """Helper function to return queue object."""
         if self._queue is None:
             self._queue = self.env.sqs_resource.get_queue_by_name(
                 QueueName=self.get_sqs_queue_name()
@@ -224,9 +281,7 @@ class RawQueue(GenericQueue):
         return func
 
     def connect_raw_processor(self, processor):
-        """
-        Assign raw processor (a function) to handle jobs.
-        """
+        """Assign raw processor (a function) to handle jobs."""
         extra = {
             "queue_name": self.name,
             "processor_name": processor.__module__ + "." + processor.__name__,
@@ -246,9 +301,7 @@ class RawQueue(GenericQueue):
         deduplication_id: Optional[str] = None,
         group_id: str = DEFAULT_MESSAGE_GROUP_ID,
     ):
-        """
-        Add raw message to the queue
-        """
+        """Add raw message to the queue"""
 
         kwargs = {
             "MessageBody": message_body,
@@ -341,7 +394,9 @@ class JobQueue(GenericQueue):
     _batch_level: int = attr.ib(default=0, repr=False)
     _batched_messages: List[Dict] = attr.ib(factory=list, repr=False)
 
-    def processor(self, job_name, pass_context=False, context_var=DEFAULT_CONTEXT_VAR):
+    def processor(
+        self, job_name: str, pass_context: bool = False, context_var=DEFAULT_CONTEXT_VAR
+    ) -> Callable[[Callable[P, Any]], AsyncTask[P]]:
         """
         Decorator to assign processor to handle jobs with the name job_name
         from the queue queue_name
@@ -367,7 +422,7 @@ class JobQueue(GenericQueue):
             say_hello(name='John')
         """
 
-        def fn(processor):
+        def fn(processor: Callable[P, Any]) -> AsyncTask[P]:
             return self.connect_processor(
                 job_name, processor, pass_context, context_var
             )
@@ -375,8 +430,12 @@ class JobQueue(GenericQueue):
         return fn
 
     def connect_processor(
-        self, job_name, processor, pass_context=False, context_var=DEFAULT_CONTEXT_VAR
-    ):
+        self,
+        job_name: str,
+        processor: Callable[P, Any],
+        pass_context: bool = False,
+        context_var=DEFAULT_CONTEXT_VAR,
+    ) -> AsyncTask[P]:
         """
         Assign processor (a function) to handle jobs with the name job_name
         from the queue queue_name
@@ -402,27 +461,21 @@ class JobQueue(GenericQueue):
         )
         return AsyncTask(self, job_name, processor)
 
-    def get_processor(self, job_name):
-        """
-        Helper function to return a processor for the queue
-        """
+    def get_processor(self, job_name: str) -> Optional[Processor]:
+        """Helper function to return a processor for the queue"""
         return self.processors.get(job_name)
 
-    def open_add_batch(self):
-        """
-        Open an add batch.
-        """
+    def open_add_batch(self) -> None:
+        """Open an add batch."""
         self._batch_level += 1
 
-    def close_add_batch(self):
-        """
-        Close an add batch.
-        """
+    def close_add_batch(self) -> None:
+        """Close an add batch."""
         self._batch_level = max(0, self._batch_level - 1)
         self._flush_batch_if_needed()
 
     @contextmanager
-    def add_batch(self):
+    def add_batch(self) -> Generator[None, None, None]:
         """
         Context manager to add jobs in batch.
 
@@ -438,25 +491,67 @@ class JobQueue(GenericQueue):
         finally:
             self.close_add_batch()
 
-    def _should_flush_batch(self):
+    def _should_flush_batch(self) -> bool:
         """
         Check if a message batch should be flushed, i.e. all messages should be sent
         and removed from the internal cache.
 
-        Right now this only checks the number of messages. In the future we may
-        want to improve that by also ensuring the batch size is small enough.
+        This not only checks the number of messages, but that the total batch size has
+        not gotten too large.
         """
         max_size = SEND_BATCH_SIZE if self._batch_level > 0 else 1
+
+        # Attempt to flush if we have gotten close to the message size limit
+        if self._estimated_message_length(self._batched_messages) > MAX_MESSAGE_LENGTH:
+            list_length = len(self._batched_messages)
+            logger.info(
+                f"Flushing SQS batch on oversized message list with only {list_length} items"
+            )
+            return True
+
         return len(self._batched_messages) >= max_size
 
-    def _flush_batch_if_needed(self):
+    def _estimated_message_length(self, messages: list) -> int:
+        """
+        Return an estimate of the length of the combined message we want to send.
+
+        The length is in bytes and we add a 1K buffer to account for any extra headers.
+        """
+        return len(json.dumps(messages).encode("utf-8")) + 1024
+
+    def _flush_batch_if_needed(self) -> None:
         queue = self.get_queue()
 
         # There should be at most 1 batch to send. But just in case, prepare to
         # send more than that.
         while self._should_flush_batch():
-            msgs = self._batched_messages[:SEND_BATCH_SIZE]
-            self._batched_messages = self._batched_messages[SEND_BATCH_SIZE:]
+            send_batch_size = len(self._batched_messages)  # will be <= SEND_BATCH_SIZE
+
+            while (
+                send_batch_size > 1
+                and self._estimated_message_length(
+                    self._batched_messages[:send_batch_size]
+                )
+                > MAX_MESSAGE_LENGTH
+            ):
+                # Progressively reduce the batch size until it fits.
+                send_batch_size -= 1
+
+            # XXX: temporary logging while we check the length estimates against what
+            # the SQS service reports. Remove if still here after 2024-02-25.
+            message_length = self._estimated_message_length(
+                self._batched_messages[:send_batch_size]
+            )
+            if message_length > MAX_MESSAGE_LENGTH:
+                # We log here so we can match up with the corresponding error message
+                # from SQS. Note, SQS errors are dynamic and difficult to catch
+                # explicitly and its not worth the trouble here.
+                logger.warning(
+                    f"SQS Message is probably too long: {message_length} bytes",
+                )
+
+            msgs = self._batched_messages[:send_batch_size]
+            self._batched_messages = self._batched_messages[send_batch_size:]
 
             # Add Ids
             msg_by_id = {}
@@ -476,13 +571,13 @@ class JobQueue(GenericQueue):
 
     def add_job(
         self,
-        job_name,
-        _content_type=None,
-        _delay_seconds=None,
+        job_name: str,
+        _content_type: Optional[str] = None,
+        _delay_seconds: Optional[int] = None,
         _deduplication_id=None,
-        _group_id=None,
-        **job_kwargs
-    ):
+        _group_id: Optional[str] = None,
+        **job_kwargs,
+    ) -> Optional[str]:
         """
         Add job to the queue. The body of the job will be converted to the text
         with one of the codecs (by default it's "pickle_compat")
@@ -504,17 +599,15 @@ class JobQueue(GenericQueue):
 
     def add_raw_job(
         self,
-        job_name,
+        job_name: str,
         message_body,
         job_context,
         content_type,
         delay_seconds,
         deduplication_id,
-        group_id,
-    ):
-        """
-        Low-level function to put message to the queue
-        """
+        group_id: Optional[str],
+    ) -> Optional[str]:
+        """Low-level function to put message to the queue"""
         # if queue name ends with .fifo, then according to the AWS specs,
         # it's a FIFO queue, and requires group_id.
         # Otherwise group_id can be set to None
@@ -552,11 +645,11 @@ class JobQueue(GenericQueue):
         Return True if processing went successful
         """
         job_name = get_job_name(message)
-        processor = self.get_processor(job_name)
+        processor = self.get_processor(job_name)  # type: ignore
         if processor:
             return processor.process_message(message)
         else:
-            return self.process_message_fallback(job_name)
+            return self.process_message_fallback(job_name)  # type: ignore
 
     def process_messages(self, messages: List[Any]) -> bool:
         """
@@ -572,17 +665,17 @@ class JobQueue(GenericQueue):
         results = []
 
         for job_name, grouped_messages in messages_by_job_name.items():
-            processor = self.get_processor(job_name)
+            processor = self.get_processor(job_name)  # type: ignore
             if processor:
                 result = processor.process_messages(grouped_messages)
                 results.append(result)
             else:
-                result = self.process_message_fallback(job_name)
+                result = self.process_message_fallback(job_name)  # type: ignore
                 results.append(result)
 
         return all(results)
 
-    def process_message_fallback(self, job_name):
+    def process_message_fallback(self, job_name: str) -> Literal[False]:
         # Note: we don't set exc_info=True, since source of the error is almost
         # certainly in another code base.
         logger.error(
@@ -593,7 +686,7 @@ class JobQueue(GenericQueue):
         )
         return False
 
-    def copy_processors(self, dst_queue: "JobQueue"):
+    def copy_processors(self, dst_queue: "JobQueue") -> None:
         """
         Copy processors from self to dst_queue. Can be helpful to process d
         ead-letter queue with processors from the main queue.
